# -*- coding: utf-8 -*-
#
# Copyright 2012, 2014 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls JMD
#
# mpuls JMD is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls JMD is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls JMD. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Björn Ricks <bjoern.ricks@intevation.de>
# Roland Geider <roland.geider@intevation.de>

"""The application's Globals object"""

import os
import logging

import mpulsweb.lib.app_globals

from formed.meta.io.document import openDocument
from formed.meta.data import RuleLeaf
from formed.meta.rulegen import RuleGenerator


log = logging.getLogger(__name__)

class Globals(mpulsweb.lib.app_globals.Globals):

    def _load_formedtree_from_file(self, formed_file):
        # Build FormedTree. This tree will work as template for all cases
        log.debug("loading formedtree definition with rules for JMD-Strukrur from %r", formed_file)
        formed = openDocument(formed_file, generate_rules=False)
        self._generate_rules(formed)
        return formed

    def _generate_rules(self, meta):
        """Generate rules for JMD Respekt Coaches.

        This is mostly like the default rule generator, but it can also
        generate manual required rules for groups of boolean fields
        where at least one must be selected by the user. To recognize
        which groups these are, the code assumes that all nodes whose
        name starts with "bool-group-" are such groups.
        """
        # first, generate the new special rules for the boolean groups.
        # Collect information about which check-boxes are part of which
        # group while doing this.
        bool_group_messages = {}
        for nc in meta.name2nc.values():
            if nc.getName().startswith("bool-group-"):
                description = (u"Mindestens eine der Auswahlmöglichkeiten in "
                               u"'%s' muss ausgewählt werden"
                               % nc.get_title())
                boolnames = []
                for child in nc.getChildren():
                    name = child.getName()
                    boolnames.append(name)
                    bool_group_messages[name] = description

                expr = (" ".join("$%s 1 ==" % name for name in boolnames)
                        + " or" * (len(boolnames) - 1))
                for name in boolnames:
                    rule = RuleLeaf()
                    rule.setName("required-rule-manual-bool-rule-%s" % name)
                    rule.setMark("warning:%s" % name)
                    rule.setDescription(description)
                    rule.setValue(description)
                    rule.setExpr(expr)
                    nc.addChild(rule)

        self.bool_group_messages = bool_group_messages

        # Now generate the standard rules. Because all rules generated
        # above are manual rules, no new required rules will be
        # generated for the booleans.
        generator = RuleGenerator(meta)
        generator.generateRequiredRules()
        generator.generateDateSequenceRules()
        meta = generator.document

        meta.init()

        return meta

    def _init_case(self, config):
        """Extend the base method so that we can handle the bund-level fields.

        Before delegating to the base method, this method filters the
        jmd_bund_only attribute from the agency fields and also removes
        those fields for which this attribute is true, unless the
        ini-file says the evaluation server is a bund-level server.
        """
        fields = self.mpuls_config.get("agency", "fields")
        fields[:] = self._filter_fields(config, fields)
        return super(Globals, self)._init_case(config)

    def _filter_fields(self, config, fields):
        is_bund_eval = config.get("jmd.eval.bund", False)
        for field in fields:
            if not field.get("jmd_bund_only") or is_bund_eval:
                if "jmd_bund_only" in field:
                    del field["jmd_bund_only"]
                yield field
