# -*- coding: utf-8 -*-
#
# Copyright 2012 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls JMD
#
# mpuls JMD is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls JMD is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls JMD. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Björn Ricks <bjoern.ricks@intevation.de>
# Roland Geider <roland.geider@intevation.de>

"""The application's Globals object"""

import os
import logging

import mpulsweb.lib.app_globals

from formed.meta.io.document import openDocument
from formed.meta.rulegen import RuleGenerator

import jmdweb.model.deletion


log = logging.getLogger(__name__)


class JMDRuleGenerator(RuleGenerator):

    """JMD-specific rule generator for better rules CM-phase required fields.

    The fields required for finishing the CM-phase should not be
    considered required fields when not actually doing case-management.
    However, they must be considered required fields for running an
    finished CM, so that changing such a field once in the finished CM
    stated may automatically change the state to running CM again.
    Strictly speaking this would probably apply to any state reachable
    from running CM, but finished CM is the only such state in JMD at
    the moment.

    This rule generator overwrites the expression_for_required_rule
    method so that the rules generated for fields required for phase_3
    (finished CM) are also considered fulfilled if the current phase is
    not 2 (running CM) or 3 (finished CM).
    """

    def expression_for_required_rule(self, widget):
        expr = super(JMDRuleGenerator,
                     self).expression_for_required_rule(widget)
        required = widget.get_parsed_flags().get("required", [])
        if "phase_3" in required:
            expr += " $phase 2 != $phase 3 != and or"
        return expr



class Globals(mpulsweb.lib.app_globals.Globals):

    def __init__(self, config):
        super(Globals, self).__init__(config)

        self.deletions = jmdweb.model.deletion.parse_deletion_config(config)

    def _load_formedtree_from_file(self, formed_file):
        # Build FormedTree. This tree will work as template for all cases
        log.debug("loading formedtree definition with rules for JMD from %r",
                  formed_file)
        return openDocument(formed_file, generate_rules=True,
                            rule_generator=JMDRuleGenerator)

    def _load_form(self, config):
        """Extend inherited method: create a formed tree for imports.
        The formed tree for imports does not contain any rules that
        refer to "today" so that XML import of old files is possible
        (mpuls/issue5614).
        """
        super(Globals, self)._load_form(config)
        formed_file = self._formed_file_name(config, 'formed', 'treefile')
        meta = self._load_formedtree_from_file(formed_file)
        self._remove_today_rules(meta)
        self.formed_tree_for_imports = meta

    def _remove_today_rules(self, meta):
        """Remove the rules referring to "today" from the formed meta tree.
        The formed meta tree is modified in place.
        """
        to_remove = set()
        for rules in meta.name2rules.values():
            for rule in rules:
                if rule.getExpr().uses_today:
                    to_remove.add(rule)

        for rule in to_remove:
            rule.getParent().removeChild(rule)

        meta.init()
