# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008, 2010 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls JMD
#
# mpuls JMD is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls JMD is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls JMD. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls JMD has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>
#

import logging
import datetime
from email.mime.text import MIMEText
from subprocess import Popen, PIPE

from pylons import app_globals as g

from mpulsweb.lib.db import db
from mpulsweb.lib.security import get_unmapped_db_name
from mpulsweb.model.agencysettings import Agency


log = logging.getLogger(__name__)


class DelectionConfig(object):

    def __init__(self, interval, notification_threshold, sender, recipients):
        self.interval = interval
        self.notification_threshold = notification_threshold
        self.sender = sender
        self.recipients = recipients

    def recipient_for_fkz(self, fkz):
        for prefix, recipient in self.recipients:
            if fkz.startswith(prefix):
                return recipient
        return None


def parse_deletion_config(config):
    """Extract the deletion configuration from the ini file.
    Returns an instance of DelectionConfig.
    """
    interval = int(config.get("jmd.deletion.interval", "7"))
    threshold = int(config.get("jmd.deletion.threshold", "10"))
    sender = config.get("jmd.deletion.sender")
    recipients = []
    for line in config.get("jmd.deletion.recipients", "").splitlines():
        parts = line.split()
        if len(parts) == 0:
            continue
        if len(parts) != 2:
            raise ValueError("jmd.deletion.recipients line %r does not have"
                             " exactly 2 items" % (line,))
        recipients.append(tuple(parts))
    return DelectionConfig(interval=interval, notification_threshold=threshold,
                           sender=sender, recipients=recipients)


def count_recently_deleted():
    return db.query("""SELECT coalesce(sum(count), 0) FROM case_deletions_tbl
                        WHERE age(deleted_at) < %s;""",
                    (datetime.timedelta(g.deletions.interval),))[0]

def record_deletion(count):
    db.execute("""INSERT INTO case_deletions_tbl (count) VALUES (%s)""",
               (count,))


def send_notification_mail(count):
    agency = Agency()
    agency_id = get_unmapped_db_name()  # certnumber/db
    to_adress = g.deletions.recipient_for_fkz(agency_id)
    if to_adress is None:
        log.warning("Could not determine recipient for deletion warning for"
                    " fkz %r" % (agency_id,))
        return

    msg_values = {"jmd_name": agency.getName(),
                  "fkz": agency_id,
                  "count": count,
                  "interval": g.deletions.interval,
                  }
    message_text = (u"Der JMD %(jmd_name)s (%(fkz)s) hat in den letzten"
                    u" %(interval)d Tagen %(count)d Akten in JMD i-mpuls"
                    u" endgültig gelöscht."
                    % msg_values)
    msg = MIMEText(message_text, _charset="utf-8")
    msg['Subject'] = (u"JMD %(jmd_name)s (%(fkz)s) hat viele Akten gelöscht"
                      % msg_values)
    msg['From'] = g.deletions.sender
    msg['To'] = to_adress

    p = Popen(["/usr/sbin/sendmail", "-t"], stdin=PIPE)
    p.communicate(msg.as_string())
