# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.

"""Routes configuration

The more specific and detailed routes should be defined first so they
may take precedent over the more generic routes. For more information
refer to the routes manual at http://routes.groovie.org/docs/
"""

import os
import re
import logging

from routes import Mapper

import mpulsweb.controllers

log = logging.getLogger(__name__)


def controller_scan(directory=None):
    """Scan a directory for python files and use them as controllers"""
    log.debug('Scanning for controllers in %s' % directory)
    controllers = []
    if directory is None:
        return []

    def find_controllers(dirname, prefix=''):
        """Locate controllers in a directory"""
        controllers = []
        for fname in os.listdir(dirname):
            filename = os.path.join(dirname, fname)
            if os.path.isfile(filename) and \
                    re.match('^[^_]{1,1}.*\.py$', fname):
                controllers.append(prefix + fname[:-3])
            elif os.path.isdir(filename):
                controllers.extend(find_controllers(filename,
                                                    prefix=prefix + fname + '/')
                                   )
        return controllers

    def longest_first(fst, lst):
        """Compare the length of one string to another, shortest goes first"""
        return cmp(len(lst), len(fst))

    for dirname in directory:
        controllers.extend(find_controllers(dirname))
    controllers.sort(longest_first)
    log.debug('Found controllers: %s' % controllers)
    return controllers


class MpulsMapper(Mapper):

    def __init__(self, config):
        # use list(set(...)) to ensure a list of unique elements.
        controller_dirs = list(set([
            # The application specific controller directory:
            config['pylons.paths']['controllers'],
            # The base controller directory:
            os.path.dirname(mpulsweb.controllers.__file__),
            ]))

        super(MpulsMapper, self).__init__(
            controller_scan=controller_scan,
            directory=controller_dirs,
            always_scan=config['debug'])

        self.explicit = False
        self.make_map(config)

    def make_map(self, config):
        # The ErrorController route (handles 404/500 error pages); it should
        # likely stay at the top, ensuring it can always be resolved
        self._add_error_map(config)
        self._add_index_map(config)

        self.add_custom_map(config)

        self._add_case_overview_map(config)
        self._add_case_bundle_map(config)
        self._add_phase_map(config)
        self._add_default_controller_map(config)

    def add_custom_map(self, config):
        # should be overriden in a mpuls application to add app specific routes
        pass

    def _add_error_map(self, config):
        self.connect('/error/{action}', controller='error')
        self.connect('/error/{action}/{id}', controller='error')

    def _add_index_map(self, config):
        self.connect('/', controller=config.get('mpuls.app.indexcontroller',
                                                'mpuls'),
                     action='index')

    def _add_case_overview_map(self, config):
        # Explicit route for case overview actions because the generic
        # '/{controller}/{action}/{id}' doesn't work for them.
        self.connect('/case_overview/{action}/{confirmed}',
                     controller="case_overview")

    def _add_case_bundle_map(self, config):
        self.connect('/case_bundle/{action}/{confirmed}',
                     controller="case_bundle")

    def _add_phase_map(self, config):
        self.connect('/phase/setPhase/{id}/{value}',
                     controller="phase", action="setPhase")

    def _add_default_controller_map(self, config):
        self.connect('/{controller}', action='index')
        self.connect('/{controller}/{action}')
        self.connect('/{controller}/{action}/{id}')
        self.connect('/{controller}/{action}/{id}/{confirmed}')
        # used for dowloading files
        self.connect('/{controller}/{action}/{id}/*url')


def make_map(config):
    """Create, configure and return the routes Mapper"""
    return MpulsMapper(config)
