# -*- coding: utf-8 -*-
from StringIO import StringIO
import logging
import itertools
import zipfile
import stat
from collections import defaultdict
from datetime import datetime, date
from decimal import Decimal
from operator import methodcaller

import xlwt

from formed.meta.data import PageNode, RepeatNode, EnumeratedNode, \
    int_or_none, BoolLeaf
from formed.meta.structure import StructureWidgetCollector

from mpulsweb.lib.base import g
from mpulsweb.lib.translation import _

log = logging.getLogger(__name__)

#
# Export base classes
#


class Export(object):

    def export(self):
        raise NotImplementedError()


class Table(object):

    """A single Table for tabular output formats like XLS or CSV"""

    def __init__(self, title, columns):
        """Initialize the table.
        The parameter title is a string with the title of the table and
        will usually be taken from the page or repeat group that is
        represented by the table.  The parameter columns is a list
        defining the column names"""
        self.title = title
        self.columns = columns
        self.rows = []

    def append(self, row):
        """Append a row to the table.
        The parameter row should be a dictionary mapping column names to
        values.
        """
        self.rows.append(row)


class Column(object):

    """A Column is a data class and contains name and title."""

    def __init__(self, name, title):
        """Initialize a Column with name and title. The title is the
        description which is the column header in the XLS Export."""
        self.name = name
        self.title = title


class FieldSelector(object):

    def __init__(self, fields, pages):
        self.fields = fields
        self.pages = pages

    def accept(self, name):
        return name in self.fields

    def get_pages(self):
        return self.pages


class Container(object):

    def fill(self, tree):
        raise NotImplementedError()


class TableContainer(Container):

    """A container of multiple Table instances"""

    def __init__(self, meta_tree, selection, table_factory=Table,
                 one_table_per_page=False, fill_enumerated_nodes=False,
                 one_sheet_export=False, squash_lines=False, get_title=None):
        """Initialize the container with a FormEd meta tree and the selection.

        The get_title parameter, if given, should be a callable which
        when called with a formed meta node returns the title to use for
        the page or column corresponding to the node. If omitted, it
        defaults to effectively calling the node's getDescription
        method. As an example, if you want to use the title instead of
        the description, you could pass
        operator.methodcaller("get_title") as the get_title parameter.
        """
        self.meta_tree = meta_tree
        self.table_factory = table_factory
        self.tables = dict()
        self.field_name_to_table_name = dict()
        self.table_order = []
        self.one_table_per_page = one_table_per_page
        self.fill_enumerated_nodes = fill_enumerated_nodes
        self.one_sheet_export = one_sheet_export
        self.squash_lines = squash_lines

        if get_title is None:
            get_title = methodcaller("get_title")
        self.create_tables(selection, get_title)

    def create_tables(self, selection, get_title):
        table_columns = dict()
        repeat_groups = dict()
        titles = dict()
        table_order = []
        for page in self.meta_tree.findAllByClass(PageNode):
            # always export the invisible pages as they cannot be
            # explicily selected by the user and they usually contain
            # important information like the current phase.
            if (not isinstance(selection, FieldSelector) and
                    page.getName() not in selection and not page.isInvisible()):
                continue

            collector = StructureWidgetCollector()
            page.visit(collector.visitor)
            for nc in collector.widgets:
                if isinstance(selection, FieldSelector) and \
                        not selection.accept(nc.getName()):
                    continue
                if not self.one_sheet_export and self.one_table_per_page:
                    table_name = page.getName()
                    table_title = get_title(page)
                else:
                    table_name = "master"
                    table_title = u"Master"
                parent_name = "master"
                for p in nc.pathToRoot():
                    if isinstance(p, RepeatNode):
                        if not self.one_sheet_export:
                            table_name = p.getName()
                            table_title = get_title(p)
                            repeat_groups[table_name] = parent_name
                            parent_name = table_name
                        else:
                            repeat_groups[p.getName()] = parent_name

                if table_name not in table_columns:
                    table_order.append(table_name)

                cols = table_columns.setdefault(table_name, [])
                if not isinstance(nc, RepeatNode):
                    cols.append(Column(nc.getName(), get_title(nc)))
                if not table_title:
                    log.warning("Table title for table %s is empty. Using "
                                " table name as title." % table_name)
                    table_title = table_name
                titles[table_name] = table_title

        for name, columns in table_columns.iteritems():
            extra_columns = []
            if name in repeat_groups and not self.one_sheet_export:
                extra_columns.append(Column(name + "_id", name + "_id"))
                if repeat_groups[name] == "master":
                    extra_columns.append(Column("master_id", "ID"))
                else:
                    extra_columns.append(Column(repeat_groups[name] + "_id",
                                                repeat_groups[name] + "_id"))
            elif name in repeat_groups:
                pass
            else:
                extra_columns.append(Column("master_id", "ID"))
            self.tables[name] = self.table_factory(titles[name],
                                                   extra_columns + columns)
            for col in columns:
                self.field_name_to_table_name[col.name] = name

        self.table_order = table_order

    def fill(self, tree):
        """Fill the tables with the contents of the ElementTree tree."""
        self.repeat_group_ids = itertools.count()
        for case_id, case in enumerate(tree.getiterator('master')):
            use_mid = self.use_master_id()
            log.debug("Export should use master id %s" % use_mid)
            if use_mid:
                # use master id if set on the etree.Element or otherwise
                # enumerated value for node_id
                case_id = case.get("id", case_id)
            result = self.fill_from_node(case, case_id)
            for table_name, row in result:
                if table_name in self.tables:
                    self.tables[table_name].append(row)

    def use_master_id(self):
        """Returns True if row should try to use master id"""
        return g.mpuls_config.is_enabled("exports", "use-master-id")

    def fill_from_node(self, node, node_id, parent_name=None, parent_id=None):
        assert parent_name is None or parent_id is not None

        rows = defaultdict(dict)
        result_rg = []

        if parent_name is not None:
            rows[node.tag][parent_name + "_id"] = parent_id

        for child in node.getchildren():
            name = child.tag
            nc = self.meta_tree.findByName(name)
            if isinstance(nc, RepeatNode):
                result_rg.extend(self.fill_from_node(
                    child, self.repeat_group_ids.next(), node.tag, node_id))
            else:
                table_name = self.field_name_to_table_name.get(name)
                if table_name is not None:
                    rows[table_name][name] = self.expand_enums(nc, child.text)

        id_column_name = node.tag + "_id"
        for row in rows.itervalues():
            row[id_column_name] = node_id

        if self.squash_lines:
            rgs = []
            for table_name, rg in result_rg:
                table_row = rows[table_name]
                rg_cols = set(rg.keys())
                if set(table_row.keys()).isdisjoint(rg_cols):
                    table_row.update(rg)
                else:
                    for tn, subrow in rgs:
                        if tn != table_name:
                            # Only merge subrows in the same table.
                            # Without this check on a single
                            # page export rows would be merged into
                            # the repeat group tables.
                            continue
                        if set(subrow.keys()).isdisjoint(rg_cols):
                            subrow.update(rg)
                            break
                    else:
                        rgs.append((table_name, rg))
        else:
            rgs = result_rg

        return rows.items() + rgs

    def expand_enums(self, nc, value):
        """Return the description for value if nc is an enumeration.
        Otherwise value will be returned unchanged."""
        if not self.fill_enumerated_nodes:
            return value

        if isinstance(nc, EnumeratedNode):
            value = nc.getValueMapping().get(int_or_none(value), value)
        elif isinstance(nc, BoolLeaf):
            # This boolean expansion is not quite right because of the
            # way false is represented in most cases: It's often stored
            # as NULL in the database and thus not present at all in the
            # XML this export is based on (see mpuls/issue5077).
            if value == "-1":
                value = u"Keine Angabe"
            elif value == "0":
                value = u'Nein'
            elif value == "1":
                value = u'Ja'
        return value


#
# XLS Export
#


default_style = xlwt.XFStyle()

datetime_style = xlwt.XFStyle()
datetime_style.num_format_str = "DD.MM.YYYY"

date_style = xlwt.XFStyle()
date_style.num_format_str = "DD.MM.YYYY"

float_style = xlwt.XFStyle()
float_style.num_format_str = "0.00"


class XLSTable(Table):

    def add_to_workbook(self, workbook):
        sheet = workbook.add_sheet(self.title)
        for columnnr, column in enumerate(self.columns):
            sheet.write(0, columnnr, column.title)

        for row_num, row in enumerate(self.rows):
            for columnnr, column in enumerate(self.columns):
                name = column.name
                value, style = self.encode(row.get(name))
                sheet.write(row_num + 1, columnnr, value, style)

    def encode(self, value):
        style = default_style
        if value is None:
            return ("", style)
        if isinstance(value, datetime):
            return (value, datetime_style)
        if isinstance(value, date):
            value = datetime(value.year, value.month, value.day)
            return (value, date_style)
        if isinstance(value, basestring) and len(value) > 32765:
            value = value[:32765]
        if isinstance(value, Decimal):
            return (float(value), style)
        if not isinstance(value, (basestring, int, long, float)):
            return (str(value), style)
        return (value, style)


class XLSContainer(TableContainer):

    sheet_name_forbidden_characters = dict((ord(c), None) for c in ":\/?*[]")

    max_sheet_name_len = 31

    def write_xls(self, output):
        used_names = set()
        workbook = xlwt.Workbook()
        for name in self.table_order:
            if name in self.tables:
                table = self.tables[name]
                table.title = self.sheet_name_from_title(table.title,
                                                         used_names)
                used_names.add(table.title)
                table.add_to_workbook(workbook)
        workbook.save(output)

    def sheet_name_from_title(self, title, used):
        """Return the name of a worksheet based on the title.
        Worksheet names have to follow the following rules (Excel 2003):

          - They may be at most 31 characters long
          - They must not contain any of the these characters: : \ / ? * [ ]
          - They must not be empty
          - They must be unique in the workbook

        The worksheet name is simply the title, unless one the rules is
        violated.  If a rule is violated, the method logs the
        appropriate warning and tries to correct the problems.  Any
        invalid characters are removed.  Names that are too long are
        shortened.  Empty or duplicate names are replaced with a generic
        but unique name.
        """
        name = title.translate(self.sheet_name_forbidden_characters)
        if name != title:
            log.warning(
                "Stripped invalid characters from sheet name %r", title)
        if len(name) > self.max_sheet_name_len:
            log.warning("excel sheet name %r is too long; stripping extra"
                        " characters", name)
            # name may end with whitespace after truncation, so strip that,
            # too.
            name = name[:self.max_sheet_name_len].strip()
        if not name or name in used:
            if not name:
                log.warning("empty name for an excel sheet; selecting unique"
                            " alternative")
                name = _("Table")
            else:
                log.warning("sheet name %r already used; selecting unique"
                            " alternative", name)
            # start counting at 1
            num = len(used) + 1
            while 1:
                numstr = " %d" % num
                candidate = (name[:self.max_sheet_name_len - len(numstr)]
                             + numstr)
                if candidate not in used:
                    name = candidate
                    break
                num += 1
        if name != title:
            log.warning("Using %r instead of %r as excel sheet name", name,
                        title)

        return name


class XLSExport(Export):

    def __init__(self, tree, selection, one_sheet_export=False,
                 squash_lines=False, get_title=None):
        self.tree = tree
        self.xls = XLSContainer(g.formedTree, selection, XLSTable,
                                one_table_per_page=True,
                                one_sheet_export=one_sheet_export,
                                squash_lines=squash_lines,
                                get_title=None)

    def export(self):
        self.xls.fill(self.tree)
        out = StringIO()
        self.xls.write_xls(out)
        contents = out.getvalue()
        out.close()
        return contents


class EnrichedXLSExport(XLSExport):

    '''Use the fill_enumerated_node to fill all EnumeratedNode_s with
       its value and not with its key'''

    def __init__(self, tree, selection, one_sheet_export=False,
                 squash_lines=False, get_title=None):
        self.tree = tree
        self.xls = XLSContainer(g.formedTree, selection, XLSTable,
                                one_table_per_page=True,
                                fill_enumerated_nodes=True,
                                one_sheet_export=one_sheet_export,
                                squash_lines=squash_lines,
                                get_title=get_title)

#
# Formletter Export
#


class FormletterXLSTable(XLSTable):

    def encode(self, value):
        if isinstance(value, (float, Decimal)):
            style = float_style
            value = float(value)
        else:
            value, style = super(FormletterXLSTable, self).encode(value)
        return (value, style)


class FormletterExport(Export):

    def __init__(self, data, order=None):
        self.data = data
        if order:
            self.order = order
        else:
            self.order = data[0].keys()

        columns = [Column(name, name) for name in self.order]
        self.xls = FormletterXLSTable('formletter', columns)

    def export(self):
        for row in self.data:
            self.xls.append(row)
        out = StringIO()
        self.write_xls(out)
        contents = out.getvalue()
        out.close()
        return contents

    def write_xls(self, output):
        workbook = xlwt.Workbook()
        self.xls.add_to_workbook(workbook)
        workbook.save(output)

#
# CSV Export
#


class CSVTable(Table):

    """A single Table that can be serialized as CSV"""

    def serialize(self, sep):
        """Return the table as a string in CSV format.

        The parameter sep is the separator character.  Rows are
        separated by newlines.  If the separator character occurs in
        values, all occurrences are replaced by spaces or, if the
        separator is the space character, by tabs.
        """
        colnames = [col.name for col in self.columns]
        lines = ["#" + sep.join(colnames)]
        for row in self.rows:
            lines.append(sep.join(self.encode(row.get(col), sep)
                                  for col in colnames))
        lines.append("")
        return "\n".join(lines)

    def encode(self, value, sep):
        # first, convert non-strings to strings
        if isinstance(value, int):
            # Simply converting ints to unicode assumes that the
            # separator char is neither a digit nor the minus sign.
            value = unicode(value)
        elif value is None:
            value = ""

        # convert strings if they contain the separator or CR or NL.
        if isinstance(value, basestring):
            rep = "\t" if sep == " " else " "
            value = value.replace(sep, rep)
            value = value.replace("\r\n", rep)
            value = value.replace("\r", rep)
            value = value.replace("\n", rep)
            return value
        else:
            raise TypeError(
                "Cannot handle type %r in CSV export" % type(value))


class CSVContainer(TableContainer):

    """A container of multiple CSVTable instances"""

    def write_zip_file(self, output, encoding="utf-8"):
        """Write the contents of the container to output as a zip-File"""
        def add(name, contents):
            name = name.encode(encoding)
            z.writestr(name, contents.encode(encoding))

            # By default all permission bits are unset when writestr is
            # used so that when unzipping the resulting zip-file, all
            # files are unreadable.  Set default permissions so that the
            # owner may read and write and all other can read.
            info = z.getinfo(name)
            # expression to set unix permissions on external_attr taken
            # from the Python 2.5 zipfile module
            permissions = (stat.S_IRUSR | stat.S_IWUSR
                           | stat.S_IRGRP | stat.S_IROTH)
            info.external_attr = (permissions & 0xFFFF) << 16L

        z = zipfile.ZipFile(output, "w", zipfile.ZIP_DEFLATED)
        sep = "\t"
        add("master.csv", self.tables["master"].serialize(sep))
        for name, table in self.tables.iteritems():
            if name != "master":
                add(name + '.csv', table.serialize(sep))


class CSVExport(Export):

    def __init__(self, tree, selection, one_sheet_export=False,
                 squash_lines=False, get_title=None):
        self.tree = tree
        self.csv = CSVContainer(g.formedTree, selection, CSVTable,
                                one_sheet_export=one_sheet_export,
                                squash_lines=squash_lines,
                                get_title=get_title)

    def export(self):
        self.csv.fill(self.tree)
        s = StringIO()
        self.csv.write_zip_file(s)
        return s.getvalue()
