# -*- coding: utf-8 -*-
#
# Copyright (c) 2007, 2014 by Intevation GmbH
#
# FormEd is free software under the terms of the
# GNU General Public License Version 3. See file
# LICENSE coming with the source of FormEd for details.
#
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>
# Björn Ricks <bjoern.ricks@intevation.de>
#

import codecs
from cgi import escape

from formed.meta.exprtree import OperatorBinary, OperatorUnary, And, Or, \
     Least, Greatest, Coalesce, Today, Const, DateNode, Var, List, RGMap, String
import formed.meta.data as data
import formed.export.sql


def sql_today(node, document):
    return "now()"

def sql_date(node, document):
    return "'%s'::date" % node.eval(None)

def sql_var(node, document):
    return node.var_name

def sql_str(node, document):
    return "'%s'" % formed.export.sql.sqlQuote(str(node))

def sql_const(node, document):
    return str(node.const)

def sql_And(node, document):
    return ("(%s AND %s)" % (recursive_parse(node.a, document),
                             recursive_parse(node.b, document)))

def sql_Least(node, document):
    return ("(LEAST(%s, %s))" % (recursive_parse(node.a, document),
                                 recursive_parse(node.b, document)))

def sql_Greatest(node, document):
    return ("(GREATEST(%s, %s))" % (recursive_parse(node.a, document),
                                    recursive_parse(node.b, document)))

def sql_Or(node, document):
    return ("(%s OR %s)" % (recursive_parse(node.a, document),
                            recursive_parse(node.b, document)))

def sql_sum(node, document):
    out = []
    if isinstance(node.a, List):
        out.append("(")
        out.append(" + ".join(recursive_parse(i, document) for i in
                   node.a.items))
    elif isinstance(node.a, RGMap):
        out.append("(SELECT SUM(")
        out.append(recursive_parse(node.a.b, document))
        out.append(")")
        out.append("FROM")
        out.append("%s_tbl" % str(node.a.a))
        out.append("rg")
        out.append("WHERE rg.master_id = master_tbl.id")
    else:
        raise ValueError("Invalid node type '%s' for sum expression %r" %
                         (type(node.a), node.a))
    out.append(")")
    return " ".join(out)

def sql_list(node, document):
    out = []
    out.append("(")
    out.append("%s" % ",".join(recursive_parse(i, document) for i in node.items))
    out.append(")")
    return " ".join(out)

def sql_rgmap(node, document):
    out = []
    out.append("(SELECT")
    out.append(recursive_parse(node.b, document))
    out.append("FROM")
    out.append("%s_tbl" % str(node.a))
    out.append("rg")
    out.append("WHERE rg.master_id = master_tbl.id)")
    return " ".join(out)

def sql_coalesce(node, document):
    out = []
    out.append("COALESCE(")
    out.append(recursive_parse(node.a, document))
    out.append(",")
    out.append(recursive_parse(node.b, document))
    out.append(")")
    return " ".join(out)

binary_operator_map = {
    "GT": ">",
    "LT": "<",
    "GE": ">=",
    "LE": "<=",
    "EQ": "=",
    "NE": "!=",
    "ADD": "+",
    "MINUS": "-",
    "MUL": "*",
    "DAYS": "-",
    "YEARS": "-",
    "CONTAINS": "IN",
    }

def sql_OperatorBinary(node, document):
    operator = binary_operator_map.get(node.operator.__name__)

    if operator is None:
        raise Exception('Cannot convert operator %s' % node)

    template = "(%s %s %s)"
    if node.operator.__name__ == "DAYS":
        template = "abs(%s %s %s)"
    elif node.operator.__name__ == "YEARS":
        template = "abs(%s %s %s)/365"
    elif node.operator.__name__ == "IN":
        template = "(%s %s ( %s ))"

    return template % (recursive_parse(node.a, document),
                       operator, recursive_parse(node.b, document))

def sql_OperatorUnary(node, document):
    out = []
    if node.operator.__name__ == "NOT":
        out.append("NOT")
        out.append("(")
        out.append(recursive_parse(node.a, document))
        out.append(")")
    elif node.operator.__name__ == "ISSET":
        out.append("(")
        out.append("%s IS NOT NULL" % recursive_parse(node.a, document))
        out.append(")")
    elif node.operator.__name__ == "KNOWN":
        if not isinstance(node.a, Var):
            raise Exception('Can not parse %s' % node)
        n = document.findByName(node.a.var_name)
        if isinstance(n, data.DateLeaf):
            ukv = "'0001-01-01'"
        elif isinstance(n, data.IntLeaf):
            ukv = '-9999999'
        elif isinstance(n, data.TextLeaf):
            ukv = "'unbekannt'"
        out.append("(%s != %s)" % (recursive_parse(node.a, document), ukv))
    elif node.operator.__name__ == "list_sum":
        out.append(sql_sum(node, document))
    else:
        print "Unkown OperatorUnary %s" % node.operator.__name__
        out.append(recursive_parse(node.a, document))
    return " ".join(out)

def recursive_parse(node, document):
    out = []
    if isinstance(node, OperatorUnary):
        out.append(sql_OperatorUnary(node, document))
    elif isinstance(node, OperatorBinary):
        out.append(sql_OperatorBinary(node, document))
    elif isinstance(node, And):
        out.append(sql_And(node, document))
    elif isinstance(node, Or):
        out.append(sql_Or(node, document))
    elif isinstance(node, Least):
        out.append(sql_Least(node, document))
    elif isinstance(node, Greatest):
        out.append(sql_Greatest(node, document))
    elif isinstance(node, Coalesce):
        out.append(sql_coalesce(node, document))
    elif isinstance(node, Var):
        out.append(sql_var(node, document))
    elif isinstance(node, Const):
        out.append(sql_const(node, document))
    elif isinstance(node, Today):
        out.append(sql_today(node, document))
    elif isinstance(node, DateNode):
        out.append(sql_date(node, document))
    elif isinstance(node, List):
        out.append(sql_list(node, document))
    elif isinstance(node, RGMap):
        out.append(sql_rgmap(node, document))
    elif isinstance(node, String):
        out.append(sql_str(node, document))
    else:
        raise Exception('Can not parse %s type(%s)' % (node, type(node)))
    return " ".join(out)

def generate_analyse(rules, document):
    out = []
    # 1. Header
    out.append('#!/bin/bash')
    out.append("DATABASES=`psql -t -c"
               " \"SELECT datname FROM pg_database WHERE datname LIKE 'ka_%';\""
               " postgres`")
    out.append('echo "DB,%s"'
               % ",".join(["%s" % str(i) for i in range(0,len(rules))]))
    out.append("for DB in ${DATABASES}")
    out.append("do")

    # 2. Querys
    known_rules = []
    command = 'q%s=`psql -t -c "%s" ${DB}`'
    for num, rule in enumerate(rules):
        rule_str = recursive_parse(rule.getExpr().prog, document)
        out.append('# Eval %s' % num)
        out.append('# %s' % rule.getDescription())
        out.append(command
                   % (num, ('select count(id) from master_tbl where not (%s)'
                            % rule_str)))
        out.append('\n')
        known_rules.append(rule_str)
    # 3. Footer
    out.append('echo "${DB},%s"'
               % ",".join(["${q%s}" % str(i) for i in range(0, len(rules))]))
    out.append("done")
    return "\n".join(out)

def generate_check(tag, name, desc, add_sql, del_sql):
    out = []
    base_sql = "SELECT id from ( %%(table)s ) m WHERE ( %(where)s )"
    out.append('<check tagid="%s">' % tag)
    out.append('<name>')
    out.append(name)
    out.append('</name>')
    out.append('<desc>')
    out.append(desc)
    out.append('</desc>')
    out.append('<addtag>')
    out.append('<sql>')
    out.append(escape(add_sql))
    out.append('</sql>')
    out.append('</addtag>')
    out.append('<deltag>')
    out.append('<sql>')
    out.append(escape(del_sql))
    out.append('</sql>')
    out.append('</deltag>')
    out.append('</check>')
    return "".join(out)

def generate_tagging(rules, document):
    out = []
    date_rules = []
    base_sql = "SELECT id from ( %%(table)s ) m WHERE ( %(where)s )"
    # 1. Checks
    out.append('<checks>')
    for num, rule in enumerate(rules):
        raw_tag = rule.getTag()

        if raw_tag == "none":
            continue

        if not raw_tag:
            tag = 1000 + num
        else:
            try:
                tag = int(raw_tag)
            except ValueError:
                raise ValueError("Rule %r has invalid tag value %r"
                                 % (rule.getName(), raw_tag,))

        name = rule.getDescription().replace("'", '"')
        desc = rule.getValue().replace("'", '"')
        add_sql = base_sql % {
            'where': "NOT ( %s )" % recursive_parse(rule.getExpr().prog,
                                                    document)}
        del_sql = base_sql % {
            'where': "TRUE"}
        if tag >= 1000:
            date_rules.append(tag)
        out.append(generate_check(tag, name, desc, add_sql, del_sql))
    # Added check for test if all date-sequence rules are ok
    tag = "10"
    name = "Datumsangaben konsistent"
    desc = ("Alle Datumsangaben in der Fallakte befinden sich in"
            " chronologisch korrekter Reihenfolge und liegen nicht in der"
            " Zukunft")
    add_sql = ("select id from ( %%(table)s ) m where not "
               "ARRAY(SELECT tag_id FROM nm_tags_cases_view WHERE case_id=m.id)"
               " && ARRAY[%s]" % ",".join([str(id) for id in date_rules]))
    del_sql = base_sql % {'where': "TRUE"}
    out.append(generate_check(tag, name, desc, add_sql, del_sql))
    out.append('</checks>')
    out.append('<!--\n')
    out.append('\n-->')
    return "".join(out)


class ExportRulesForTagging:

    def __init__(self, document):
        self.document = document

    @staticmethod
    def getDescription():
        return "Exports rules for Tagging as a XML"

    def doExport(self, path):
        root = self.document.root
        document = self.document
        out = None
        sql = None
        try:
            out = codecs.open(path, "wb", "UTF-8")
            all_rules = list(document.findAllByClass(data.RuleLeaf))
            error_rules = []
            known_rules = []
            rg_fields = []
            for idx, rule in enumerate(all_rules):
                ignore_rule = False

                # If rule does not result in error, ignore
                if rule.getMark().find('error') < 0:
                    ignore_rule = True

                # If rule already generated, ignore
                rule_str = recursive_parse(rule.getExpr().prog, document)
                if rule_str in known_rules:
                    ignore_rule = True

                # If field is in in a repeat group, ignore
                # TODO: if this is needed, the SQL statements have to be adapted
                for dependent_field in rule.getExpr().getDependencies():
                    field = document.findByName(dependent_field)
                    if field in rg_fields:
                        ignore_rule = True
                    for parent in field.pathToRoot():
                        if isinstance(parent, data.RepeatNode):
                            rg_fields.append(field)
                            ignore_rule = True

                if ignore_rule:
                    continue

                error_rules.append(rule)
                known_rules.append(rule_str)
            sql = generate_tagging(error_rules, document)
            print >> out, sql
        finally:
            if out:
                try:
                    out.close()
                except:
                    pass

class ExportSQLForVirtualFields:
    def __init__(self, document):
        self.document = document
        self.all_rules = list(self.document.findAllByClass(data.VirtualLeaf))


    @staticmethod
    def getDescription():
        return "Exports SQL generated from expr of VirtualFields"

    def doExport(self, fieldname):
        sql = None
        rg_fields = []
        for idx, rule in enumerate(self.all_rules):
            if rule.getName() != fieldname: continue
            ignore_rule = False
            sql = recursive_parse(rule.getExpr().prog, self.document)
            # If field is in in a repeat group, ignore
            # TODO: if this is needed, the SQL statements have to be adapted
            for dependent_field in rule.getExpr().getDependencies():
                field = self.document.findByName(dependent_field)
                if field in rg_fields:
                    ignore_rule = True
                for parent in field.pathToRoot():
                    if isinstance(parent, data.RepeatNode):
                        rg_fields.append(field)
                        ignore_rule = True

            # TODO handle rgs correctly
            # if ignore_rule:
            #     return None
            return sql


class ExportRulesAsSH:

    def __init__(self, document):
        self.document = document

    @staticmethod
    def getDescription():
        return "Export Rules As SH Analyse script"

    def doExport(self, path):
        document = self.document
        root = self.document.root
        out = None
        sql = None
        try:
            out = codecs.open(path, "wb", "UTF-8")
            all_rules = list(document.findAllByClass(data.RuleLeaf))
            error_rules = []
            known_rules = []
            for idx, rule in enumerate(all_rules):
                if rule.getMark().find('error') < 0:
                    continue
                # filter out already generated rules.
                rule_str = recursive_parse(rule.getExpr().prog, document)
                if rule_str in known_rules:
                    continue
                error_rules.append(rule)
                known_rules.append(rule_str)
            sql = generate_analyse(error_rules, document)
            print >> out, sql
        finally:
            if out:
                try:
                    out.close()
                except:
                    pass

# vim:set ts=4 sw=4 si et sta sts=4 enc=utf-8 fenc=utf-8:
