# -*- coding: latin-1 -*-
#
# Copyright 2019 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.

"""
Analyse conditionals and prepare information for their client side handling.

This does not support everything that conditionals can be used for. Only
what is necessary for JMD has been implemented. In most (although
perhaps not all) cases when a formedtree.xml file uses something that is
not supported, the code in this module will raise an exception, so that
it's clear when unsupported features are used.
"""

from collections import defaultdict
import json

import formed.meta.exprtree as E
import formed.meta.data as data


class UnsupportedFeatureError(Exception):

    """Raised when an unsupported formed tree feature is found.

    Not all feature supported by libformed were implemented initially
    because that was not necessary. When this exception is raised, it's
    an indication that perhaps support should be added.
    """


class UnknownFieldTypeException(UnsupportedFeatureError):

    """Raised when a field type encountered that cannot yet be handled.

    Not all field types were implemented initially because that was not
    necessary. When this exception is raised, it's an indication that
    perhaps support should be added.
    """


def rg_ancestors(nc):
    """Return a list of the repeat groups contain the given nc"""
    return [item for item in reversed(nc.pathToRoot())
            if isinstance(item, data.RepeatNode)]

def has_invisible_ancestors(nc):
    """Return whether any of the group or page nodes containing nc are invisible
    """
    return any(node.isInvisible()
               for node in nc.pathToRoot()
               if isinstance(node, (data.GroupNode, data.PageNode)))


class ExprSexp(object):

    def __init__(self, expr):
        self.sexp = self.expr_sexp(expr)

    def expr_sexp(self, expr):
        if isinstance(expr, E.OperatorBinary):
            return self.expr_sexp_binary_op(expr)
        elif isinstance(expr, E.Or):
            return self.expr_sexp_binary("Or", expr)
        elif isinstance(expr, E.And):
            return self.expr_sexp_binary("And", expr)
        elif isinstance(expr, E.Coalesce):
            return self.expr_sexp_binary("Coalesce", expr)
        elif isinstance(expr, E.OperatorUnary):
            return self.expr_sexp_unary_op(expr)
        elif isinstance(expr, E.Var):
            return self.expr_sexp_var(expr)
        elif isinstance(expr, E.Const):
            return self.expr_sexp_const(expr)
        elif isinstance(expr, E.String):
            return self.expr_sexp_string(expr)
        elif isinstance(expr, E.List):
            return self.expr_sexp_list(expr)
        elif isinstance(expr, E.DateNode):
            return self.expr_sexp_date(expr)
        else:
            raise UnsupportedFeatureError("Unsupported expression type: %r"
                                          % (expr,))

    _binary_operators = {
        E.EQ: "EQ",
        E.NE: "NE",
        E.GT: "GT",
        E.GE: "GE",
        E.LT: "LT",
        E.LE: "LE",
        E.CONTAINS: "CONTAINS"
        }

    def expr_sexp_binary_op(self, expr):
        known = self._binary_operators.get(expr.operator)
        if known:
            return self.expr_sexp_binary(known, expr)
        else:
            # TODO: handle more cases
            raise UnsupportedFeatureError("Unsupported binary operator: %r"
                                          % (expr.operator,))

    def expr_sexp_binary(self, opname, expr):
        return [opname, self.expr_sexp(expr.a), self.expr_sexp(expr.b)]

    _unary_operators = {
        E.NOT: "NOT",
        E.ISSET: "ISSET",
        E.KNOWN: "KNOWN",
        }

    def expr_sexp_unary_op(self, expr):
        known = self._unary_operators.get(expr.operator)
        if known:
            return self.expr_sexp_unary(known, expr)
        else:
            # TODO: handle more cases
            raise UnsupportedFeatureError("Unsupported unary operator: %r"
                                          % (expr.operator,))

    def expr_sexp_unary(self, opname, expr):
        return [opname, self.expr_sexp(expr.a)]

    def expr_sexp_var(self, expr):
        return ["Field", expr.var_name]

    def expr_sexp_const(self, expr):
        if isinstance(expr.const, int):
            return ["Const", expr.const]
        else:
            raise UnsupportedFeatureError("Unsupported constant type: %s"
                                          " (value %r)"
                                          % (type(expr.const), expr.const))

    def expr_sexp_string(self, expr):
        return ["Const", expr.str]

    def expr_sexp_list(self, expr):
        return ["List"] + [self.expr_sexp(e) for e in expr.items]

    def expr_sexp_date(self, expr):
        return ["Date", self.expr_sexp(expr.day),
                self.expr_sexp(expr.month), self.expr_sexp(expr.year)]



class CondInfo(object):

    """Information about a conditional.

    Attributes:
       name (string): the name of the conditional's meta node

       affected_fields (list of string): The names of fields affected by
           the conditional, i.e. the ones whose visibility or
           editability is affected.

       cond: The conditional node component

       expr: The expression that needs to be evaluated on the client
           side. This is the expression of the conditional together with
           expressions for the error rules of the fields that are
           referenced by the conditional expression.

       deps (set of string): The names of all the fields that are
           referenced by the expression in expr.

       groups (list of string): The names of the groups containing
           fields whose visibility is controlled by the conditional. The
           whole group will be hidden/shown on the client side because
           the group is not just the actual field but usually a
           table-element containing the input element
    """

    def __init__(self, name, affected_fields, cond, expr, deps, groups):
        self.name = name
        self.affected_fields = affected_fields
        self.cond = cond
        self.expr = expr
        self.deps = deps
        self.groups = groups


# Map field meta types to converter names used on the client side.
FIELD_TYPE_CONVERTERS = {
    data.ChoiceNode: "int",
    data.BoolLeaf: "bool",
    data.RadioNode: "radio",
    data.DateLeaf: "date",
}


class FieldInfo(object):

    """Information about a field

    Attributes:
        name (string): The name of the field's meta node.
        converter (string): The name of the converter to use on the
            client side.
        rgnames (list of string): The names of the repeat groups
            containing the field
    """

    def __init__(self, name, converter, rgnames):
        self.name = name
        self.converter = converter
        self.rgnames = rgnames


class PageInfo(object):

    """Information about a form page

    Attributes:
        name (string): The name of the page's meta node.
        conditionals (dict mapping names to CondInfo): The conditionals
            on the page which need to be handled by the client
        foreign_fields (set of string): The names of the fields that
            are referenced by conditionals on the page which are not
            located on that form page and therefore have to be added to
            the form as hidden input fields so that they're available to
            the client.
    """

    def __init__(self, name, conditionals, foreign_fields):
        self.name = name
        self.conditionals = conditionals
        self.foreign_fields = foreign_fields


class FormInfo(object):

    """Information about the form

    Attributes:
        document: The document object (i.e. the meta tree)

        pages (dict mapping names to PageInfo): Information about the
            pages of the document.

        fields (dict mapping names to FieldInfo): Information about the
            fields of the document. Only fields that are needed for
            client side conditionals are covered, though.
    """

    def __init__(self, document):
        self.document = document
        self.pages = self.build_pages()
        self.fields = self.build_fields()

    def build_pages(self):
        """Build and return the pages dictionary."""
        pages = {}
        for pagename, fields in self.document.page_fields.items():
            if self.document.name2nc[pagename].isInvisible():
                continue
            conds = defaultdict(list)
            for field in fields:
                for cond in self.document.name2conditionals[field.getName()]:
                    if has_invisible_ancestors(cond):
                        continue
                    conds[cond.getName()].append(field.getName())
            pages[pagename] = self.page_info(pagename,
                                             dict((name,
                                                   self.cond_info(name, fields))
                                                  for name, fields
                                                  in conds.items()))
        return pages

    def cond_info(self, name, fields):
        """Build and return the CondInfo object for a conditional.

        Args:
            name (string): The name of the conditional
            fields (list of string): The names of the fields affected by
                the conditional.
        """
        cond = self.document.name2nc[name]
        expanded, deps = self.expand_conditional_expr(cond.getExpr())
        groups = []
        if cond.isInvisible():
            for child in cond.getChildren():
                if isinstance(child, data.GroupNode):
                    groups.append(child.getName())
        return CondInfo(name, fields, cond, expanded, deps, groups)

    def page_info(self, name, conditionals):
        """Build and return the PageInfo object for a page.

        The conditionals are filtered before they're included in the
        PageInfo object because we're only interested in conditionals
        for which client-side evaluation is useful. This means that
        conditionals meeting the following criteria will not be included
        in the PageInfo object:

         - All fields the conditional depends on are on a different page

           In this case, the result of evaluating the expression cannot
           be influenced by the user, so it is not necessary to handle
           it on the client side

         - The conditional is invisible and the groups attribute is empty.

           The only reason for this restriction is that the code in the
           client and the renderer is simpler if we only handle the case
           when the invisible conditionals, i.e. ones whose children are
           not shown at all when the expression evaluates to false, only
           have groups as children.

        Args:
            name (string): The name of the conditional
            conditionals (dict mapping names to CondInfo): The
                conditionals contained in the page.
        """
        filtered = {}
        foreign_fields = set()
        for condname, condinfo in conditionals.items():
            foreign = set()
            same_page_deps = False
            for dep in condinfo.deps:
                dep_page = self.document.name2page[dep]
                if dep_page.getName() != name:
                    foreign.add(dep)
                else:
                    same_page_deps = True
            if not same_page_deps:
                continue
            if (self.document.name2nc[condname].isInvisible()
                and not condinfo.groups):
                continue

            foreign_fields |= foreign
            filtered[condname] = condinfo
        for field in foreign_fields:
            if isinstance(self.document.name2nc[field], data.RadioNode):
                # RadioNode is not supported here because the client
                # side needs to treat radio inputs differently in the
                # DOM and it wouldn't work correctly for hidden input
                # fields without changes.
                raise UnsupportedFeatureError("RadioNode fields are not"
                                              " supported as foreign fields"
                                              " (field: %s)." % (field,))
        return PageInfo(name, filtered, foreign_fields)

    def expand_conditional_expr(self, expr):
        """Add the error rules of all referenced fields to the expression.

        When evaluating a conditional on the server side, only values
        that conform to the error rules are considered. To replicate
        this on the client side, we extend the expression of a
        conditional to with the expressions of all the error rules that
        reference the fields referenced by the conditional, and that
        recursively for all the fields referenced by the rules that are
        added. Since all the rules must be true before the value of the
        conditional expression itself is used, the component expressions
        are joined together with "and".

        The return value is the new, extended expression and the set of
        the dependencies of the new expressions as a set of field names.
        """
        all_deps = set()
        prog = expr.prog
        todo = list(expr.depends)
        while todo:
            dep = todo.pop()
            if dep in all_deps:
                continue
            all_deps.add(dep)
            for rule in self.document.name2rules[dep]:
                if rule.isError():
                    prog = E.And(rule.getExpr().prog, prog)
                    todo.extend(rule.getExpr().getDependencies())
        return (prog, all_deps)

    def build_fields(self):
        """Build and return the dictionary for the fields attribute"""
        names = set()
        for page in self.pages.values():
            for cond in page.conditionals.values():
                names |= cond.deps
        fields = {}
        for name in names:
            nc = self.document.name2nc[name]
            converter = FIELD_TYPE_CONVERTERS.get(nc.__class__)
            if converter is None:
                raise UnknownFieldTypeException("No converter for %r" % (nc,))
            rg_names = [rg.getName() for rg in rg_ancestors(nc)]
            if len(rg_names) > 1:
                raise UnsupportedFeatureError("Nested repeat groups are not"
                                              " supported:"
                                              " field %s in rgs %s"
                                              % (name, ", ".join(rg_names)))
            fields[name] = FieldInfo(name, converter, rg_names)
        return fields

    def as_js(self):
        pages = {}
        for pagename, page in self.pages.items():
            sensitive_fields = set()
            affected_fields = defaultdict(list)
            affected_groups = defaultdict(list)
            page_info = {}
            for cond, condinfo in page.conditionals.items():
                ast = dict(expr=ExprSexp(condinfo.expr).sexp)
                page_info[cond] = ast
                sensitive_fields |= condinfo.deps
                for fieldname in condinfo.affected_fields:
                    affected_fields[fieldname].append(cond)
                for groupname in condinfo.groups:
                    affected_groups[groupname].append(cond)
            pages[pagename] = dict(conds=page_info,
                                   sensitive=sorted(sensitive_fields),
                                   affected_fields=affected_fields,
                                   affected_groups=affected_groups)

        field_info = dict((name, dict(converter=field.converter,
                                      rgs=field.rgnames))
                          for name, field in self.fields.items())

        form_info = dict(pages=pages, fields=field_info)
        return "form_info = " + json.dumps(form_info, sort_keys=True, indent=4,
                                           separators=(",",": "))
