#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2009 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

import sys
import getopt

import psycopg2 as dbapi

from formed.meta.io.document import openDocument
from formed.instance.tree import InstanceNode
from formed.instance.backends.database import DB
from formed.instance.backends.postgres import DBFactory


DEFAULT_PORT = 5432
DEFAULT_HOST = "localhost"


class PG(DB):

    def __init__(self, dbname, user, password, host, port):
        self.dbname = dbname
        self.user = user
        self.password = password.replace("\\", "\\\\")
        self.host = host
        self.port = int(port)
        self.conn = None

    def acquireCursor(self):
        if self.conn is None:
            self.conn = dbapi.connect(database=self.dbname,
                                      host=self.host, port=self.port,
                                      user=self.user, password=self.password)
        return self.conn.cursor()

    def releaseCursor(self, cursor):
        if cursor:
            try:
                cursor.close()
            except:
                pass

    def close(self):
        if self.conn is not None:
            try:
                self.conn.close()
            except:
                pass


def traverse(node, depth=0):
    print "%s%s (%s)" % ("  " * depth, node.__class__.__name__,
                         node.getMeta().getName())
    if isinstance(node, InstanceNode):
        for child in node.getChildren():
            traverse(child, depth + 1)


def usage(msg="", exit_code=0):
    print >> sys.stderr, """%s
usage:
%s -f <formed> -d <database> -u <user> -P <password> [-i <dataset id>] [-h <host>] [-p <port>]
""" % (msg, sys.argv[0])
    sys.exit(exit_code)


def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "f:d:u:p:h:P:i:",
                                   ["formed=", "database=", "user=",
                                    "password=", "host=", "port=", "id="])
    except getopt.GetoptError, err:
        usage(str(err), 1)

    port = DEFAULT_PORT
    host = DEFAULT_HOST
    database = None
    formed_file = None
    user = None
    password = None
    dataset_id = None

    for o, a in opts:
        if o in ("p", "-p", "--port"):
            try:
                port = int(a)
                if port < 0:
                    raise ValueError()
            except ValueError:
                usage("Error: port is not a postive integer", 1)
        elif o in ("h", "-h", "--host"):
            host = a
        elif o in ("d", "-d", "--database"):
            database = a
        elif o in ("u", "-u", "--user"):
            user = a
        elif o in ("P", "-P", "--password"):
            password = a
        elif o in ("f", "-f", "--formed"):
            formed_file = a
        elif o in ("i", "-i", "--id"):
            try:
                dataset_id = int(a)
                if dataset_id < 0:
                    raise ValueError()
            except ValueError:
                usage("Error: dataset id is not a postive integer", 1)

    if formed_file is None:
        usage("Error: missing formed file", 1)
    if database is None:
        usage("Error: missing database", 1)
    if user is None:
        usage("Error: missing user", 1)
    if password is None:
        usage("Error: missing password", 1)

    formed = openDocument(formed_file)

    db = None
    try:
        db = PG(database, user, password, host, port)
        factory = DBFactory(formed, db)

        if dataset_id is None:
            # create new one
            instance_tree = factory.createNewInstanceTree()
        else: # load existing one
            instance_tree = factory.loadInstanceTreeByIdentifier(dataset_id)

        traverse(instance_tree.getRootNode())
        x = instance_tree.getItem("name:%s" % dataset_id)
        print repr(x)
    finally:
        if db:
            db.close()

if __name__ == '__main__':
    main()

# vim:set ts=4 sw=4 si et sta sts=4 enc=utf-8 fenc=utf-8 :
