# -*- coding: utf-8 -*-
# Copyright (C) 2015, 2017 by Intevation GmbH
# Authors:
# Bernhard Herzog <bernhard.herzog@intevation.de>

import unittest

from formed.meta.semantic import IntChecker, TextChecker, TextAreaChecker, \
     SemanticError


class MockNodeComponent(object):

    def __init__(self, min_value=None, max_value=None, max_length=None):
        self._min_value = min_value
        self._max_value = max_value
        self._max_length = max_length

    def getMinValue(self):
        return self._min_value

    def getMaxValue(self):
        return self._max_value

    def getMaxLength(self):
        return self._max_length


class IntCheckerTest(unittest.TestCase):

    def convert_and_check(self, unparsed, min_value=None,
                          max_value=None):
        nc = MockNodeComponent(min_value=min_value, max_value=max_value)
        checker = IntChecker()
        return checker.checkAndConvert(unparsed, nc)

    def test_no_bounds(self):
        self.assertEqual(self.convert_and_check("123"), 123)

    def test_within_bounds(self):
        self.assertEqual(self.convert_and_check("534", min_value=0,
                                                max_value=1000),
                         534)

    def test_too_high(self):
        self.assertRaises(SemanticError,
                          self.convert_and_check, "10100", max_value=1000)

    def test_too_low(self):
        self.assertRaises(SemanticError,
                          self.convert_and_check, "-132", min_value=-20)

    def test_nondigits(self):
        self.assertRaises(SemanticError, self.convert_and_check, "asdf")


    def test_float_overflow(self):
        self.assertRaises(SemanticError, self.convert_and_check,
                          "99999999999999999999999e99999999999999999")


class TextCheckerMixin(object):

    def convert_and_check(self, unparsed, **kw):
        nc = MockNodeComponent(**kw)
        checker = self.checker_class()
        return checker.checkAndConvert(unparsed, nc)

    def test_simple_unicode(self):
        # unicode with 'normal' characters should not be changed.
        self.assertEqual(self.convert_and_check(u"hjkl"), u"hjkl")

    def test_utf8_decoding(self):
        # byte-strings should be utf-8 encoded.
        self.assertEqual(self.convert_and_check("aumlaut: \xc3\xa4"),
                         u"aumlaut: ä")

    def test_xml_invalid_characters(self):
        # characters not allowed in XML are filtered out.
        self.assertEqual(self.convert_and_check("This should be removed:\x01"),
                         u"This should be removed:")


class TextCheckerTest(TextCheckerMixin, unittest.TestCase):

    checker_class = TextChecker

    def test_maxlength(self):
        self.assertEqual(self.convert_and_check("01234567890123456789",
                                                max_length=10),
                         "0123456789")


class TextAreaCheckerTest(TextCheckerMixin, unittest.TestCase):

    checker_class = TextAreaChecker
